/*
 * File name : anne15.c
 * 
 * This program is an implementation of a Neural Network (NN) which attempts to
 * reach a target, by using a optimal search method based on this value using
 * Genetic  Algorithms (GA).
 *
 * It initially chooses several random input datasets (depending on the number
 * of populations)  and various predictions are made on these inputs. The 
 * error involved is used to "rank" the predictions.
 * The values with the lowest errors are seen as the "fittest" and are more 
 * likely to reproduce and get closer to the target. The process can stop after
 * the prescribed number of generations, or if the target has been reached
 * before the max. number of generations requested.
 * 
 */

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <sys/time.h>

/* NN related */
#define LIMIT     150  /* Maximum number of inputs the system can handle */
#define SESSIONS  3  /* Number of generations we'll put the system through */

/* GA related */
#define POPS        2  /* Number of populations */
#define MAXPOP     20  /* Size of population */
#define BESTPOP     3  /* Number of individuals taken from the best */
#define CROSSPOP   17  /* CROSSPOP-BESTPOP: number of individuals recombined */
#define LMUTPOP    18  /* LMUTPOP-CROSSPOP: number of individuals slightly 
			  mutated */ 
#define HMUTPOP    19  /* HMUTPOP-LMUTPOP: number of individuals hardly
			  mutated */
#define MIXPOP     10  /* Number of generations between population mixing */

int e=0;
int NUM=-1;
int SIZE=0;
char name[20][10];

/* NN related */
float target_0,target_1,target_2;
float *w1;
int total;

/* GA related */
float ***pop;
float ***temp_pop;
double score[POPS][MAXPOP];
double result[POPS][MAXPOP];
double error[POPS][MAXPOP];
float *f,*low,*high;
float *sort_low,*sort_high;
int *v;
int *x;

/*-----------------------------------------------------------------------*\
|                                                                         |
|  Randomize                                                              |
|                                                                         |
\*-----------------------------------------------------------------------*/

randomize()
{
  struct timeval tp;
  struct timezone tzp;

  /* Use time of day to feed the random number generator seed */
  gettimeofday( &tp, &tzp);
  srandom( tp.tv_sec );
}

/*-----------------------------------------------------------------------*\
|                                                                         |    
|  Get data (reads input file)                                            |
|                                                                         |
\*-----------------------------------------------------------------------*/

int get_data()
{
  FILE *fd;

  /* Total number of input values */
  total = 0;

  /* opens the file  */
  if ( (fd = fopen("nn-input","r")) == NULL )
    {
      printf ("no-input-file");
      exit(1);
    }
  else
    {    
      fscanf(fd,"%f%f",&target_0,&target_2);
      printf("Target is %f\nWanted accuracy is %f\n",target_0,target_2);
    }
  fclose( fd );
  return (0) ;
}

/*-----------------------------------------------------------------------*\
|                                                                         |
|  Return the number of cases for which the NN returns the correct value  |
|                                                                         |
\*-----------------------------------------------------------------------*/

calc_limit_err()
{
  float err=0;
  
  err = pow((target_2*target_0),2);
  target_1 = 1.0/sqrt(err) ;
  printf("Target Score is: %f\n",target_1);
  return err;
} 

/*-----------------------------------------------------------------------*\
|                                                                         |
|  Initialize the populations                                             |
|                                                                         |
\*-----------------------------------------------------------------------*/

make_initial_population()
{
  int p,i,k,y[20];
  int j,l;
  char label[60]; 
  FILE *afp;
  float depop_fe, depop_y[20],sum=0;
  
  /* read low and high values for the SIZE inputs which vary */ 
  if((afp=fopen("values","r"))==NULL)
    {
      printf("Error:can't read the file Values");
      exit(1);
    }
  else
    {
      l=0;
      fgets (label,sizeof(label),afp); 
      for (j = 0; j < NUM; j++) 
	{
	  fscanf(afp,"%s%d%d%f%f%f",name[j],&x[j],&v[j],&f[j],&low[j],&high[j]);
	  if (v[j]==1)
	    {	    
	      sort_low[l]=low[j];
	      sort_high[l]=high[j];
	      ++l;
	    }
	}
    }      
  fclose(afp);
 
  for ( p = 0 ; p < POPS ; p++ ) 
    {
      for ( i = 0 ; i < MAXPOP ; i++ ) 
	{
	  sum =0;	 
	  for ( k = 0 ; k < (SIZE-1) ; k++ )
	    {
	      y[k] = irand(SIZE-1);
	      for ( l = 0; l < k ; l++)
		{
		  while (y[k] == y[l] )
		    {
		      y[k] = irand(SIZE-1);
		      l = 0;
		    }
		}
	      depop_y[k] = (random()&1000000) / 10000.0 ;
	      sum += depop_y[k];
	    }
	  depop_fe = (random()&1000000) / 10000.0 ;
	  sum += depop_fe;
	  for ( k = 0 ; k < (SIZE-1) ; k++ )
	    {
	      j = y[k];
	      pop[p][i][j] = ((((depop_y[k])*100)/sum)-(sort_low[j]))/(sort_high[j]-sort_low[j])-0.5;
	    }
	  pop[p][i][SIZE-1] = ((random()&1048575)/1000000.0 - 0.5) * 2;
	}
    }
}

/*-----------------------------------------------------------------------*\
|                                                                         |
|  apply( p, i ) -  apply the i gene of the population p on the NN        |
|                                                                         |
\*-----------------------------------------------------------------------*/

apply( p, i )
     int p, i;
{
  int l;

  for( l = 0; l < SIZE; l++)
    {
      w1[l] = pop[p][i][l];
    }
}

/*-----------------------------------------------------------------------*\
|                                                                         |
|  This function computes the NN's output for a certain input vector.     |
|                                                                         |
\*-----------------------------------------------------------------------*/

net(p,i) 
     int p,i;     
{
  FILE *nmin,*nmin2,*ofp,*afp;
  int j,l,na,nb;
  char c,model[250][50],s[100],d[50];
  float serr=0.0,res[250],err[250],sres=0.0,sres_=0.0;
  float value_Ti,value_Nb,value_C,value_N;

  printf("Calling net with p=%d and i=%d\n",p,i);

  /* write in the file norm_test.in a random value for the inputs allowed to */
  /* vary, otherwise the normalised fixed value                              */
  if ((nmin = fopen("norm_test.in","w")) == NULL )
    {
      printf ("Can't write norm_test.in\n");
      exit(1);
    } 
  else
    {
      l = 0;
      for ( j = 0; j < NUM ; j++)
	{
	  if (v[j]==1)
	    {	   	  
	      fprintf(nmin,"%f\n",w1[l]);
	      if (!strcmp(name[j],"Ti"))	
		value_Ti = (w1[l]+0.5)*(high[j]-low[j])+low[j];
	      else
		{
		  if (!strcmp(name[j],"Nb"))	 
		    value_Nb = (w1[l]+0.5)*(high[j]-low[j])+low[j];
		  else
		    {	
		      if (!strcmp(name[j],"C"))  
			value_C = (w1[l]+0.5)*(high[j]-low[j])+low[j];
		      else
			{ 
			  if (!strcmp(name[j],"N")) 
			    value_N = (w1[l]+0.5)*(high[j]-low[j])+low[j];
			}
		    }
		}
	      l++;
	    }
	  else 
	    {
	      if (v[j]==0)
		{  
		  fprintf(nmin,"%f\n",(f[j]-low[j])/(high[j]-low[j])-0.5);
		  if (!strcmp(name[j],"Ti"))
		    value_Ti = f[j];
		  else
		    {
		      if (!strcmp(name[j],"Nb"))
			value_Nb = f[j];
		      else
			{
			  if (!strcmp(name[j],"C"))
			    value_C = f[j];
			  else
			    {		
			      if (!strcmp(name[j],"N"))
				value_N = f[j];
			    }
			}	
		    }
		}
	      else
		fprintf(nmin,"%f\n",(((((value_Ti/4)+(value_Nb/8))/(value_C+value_N))-low[j])/(high[j]-low[j]))-0.5);  
	    }	
	}
    }
  fprintf(nmin,"\n");
  fclose(nmin); 
  
  /* count the number of models */
  ofp = popen("ls _w*f","r");
  na=0;
  while((fgets(d,sizeof(d),ofp))!=NULL)  
    ++na;
  pclose(ofp);
  
  /* read files _w*f in the current directory and attribute to model[nb]*/
  ofp=popen("ls _w*f","r");
  for (nb=1 ; nb<na+1 ; nb++)
    fscanf(ofp, "%s",&model[nb]);
  pclose (ofp); 
         
  /*generate44 for each model[nb]*/
  for(nb=1; nb < na +1 ; nb++)
    {
      sprintf(s, "generate44 spec1.t1 %d %s %s.lu> /dev/null",model[nb][2]-96,&model[nb],&model[nb]);
      system(s);
      if ( (nmin2 = fopen("_out","r")) == NULL) 
	{
	  printf ("Can't read _out stage 1\n");
	  exit(1);
	}
      else
	{
	  fscanf(nmin2,"%f %f",&res[nb+1],&err[nb+1]);
	  printf("res%d err%d %f %f\n",nb+1,nb+1,res[nb+1],err[nb+1]);
	  printf("Indices %d %d\n",p,i);
    
	  fclose(nmin2);
	} 
      sres += res[nb+1];
      serr += pow(err[nb+1],2);      
      sres_ +=  pow ((res[nb+1]-target_0),2);
    }
  
 printf("thisisp %d %d\n", p,i); 

 result[p][i] = sres / na;
 error[p][i] = sqrt((serr/na)+(sres_/na)) ;

 printf("error[p][i] is %f\n\n",error[p][i]);
  
 e++;

 printf("\n-----------> sum %d\n",e);     
 printf("In net() error is %f res is %f\n\n",error[p][i],result[p][i]);
}
/*-----------------------------------------------------------------------*\
|                                                                         |
|  Return the number of cases for which the NN returns the correct value  |
|                                                                         |
\*-----------------------------------------------------------------------*/

double check_performance (p,i)
     int p,i;
{
  float score_value=0.0;
 
  net(p,i);

  score_value = error[p][i];
  printf ("Answer is %f and target is %f\n",result[p][i],target_1);

  if (score_value < 0)
    score_value = 0;
  else
    score_value = 1.0/(score_value);
      


  printf("score_value is: %f\n",score_value);
  return score_value;
}
/*-----------------------------------------------------------------------*\
|                                                                         |
|   Calculate the scores of all the vectors in all the populations        |
|                                                                         |
\*-----------------------------------------------------------------------*/

calc_score()
{
  int p, i, j;
  double check_performance();
  
  for ( p = 0 ; p < POPS ; p++ )
    {
      for ( i = 0 ; i < MAXPOP ; i++ )
	{
	  { 
	    printf("Here is scorecalc for pop %d gene %d: %f\n\n",p,i,score[p][i]);
	    apply( p, i);
	    score[p][i] = check_performance(p,i);
	    printf("Here is scorecc2 for pop %d gene %d: %f\n\n",p,i,score[p][i]);
	    printf("Read %f %f\n", result[p][i],error[p][i]);
	  }
	}
    } 
}
/*-----------------------------------------------------------------------*\
|                                                                         |
|  pop_swap( p, a, b) - swap two vectors and scores in the population p   |
|                                                                         |
\*-----------------------------------------------------------------------*/

pop_swap( p, a, b)
     int p, a, b;
{
  int i;
  float tg,t;

  /* Swap vector */
  for ( i = 0 ; i < SIZE ; i++ ) 
    {
      t = pop[p][a][i];
      pop[p][a][i] = pop[p][b][i];
      pop[p][b][i] = t;
    }
  
  /* Swap score, result ad error */
  tg = score[p][a];
  score[p][a] = score[p][b];
  score[p][b] = tg;
  
  tg = result[p][a];
  result[p][a] = result[p][b];
  result[p][b] = tg;
  
  tg = error[p][a];
  error[p][a] = error[p][b];
  error[p][b] = tg;
}
/*-----------------------------------------------------------------------*\
|                                                                         |
|   Sort the populations                                                  |
|                                                                         |
\*-----------------------------------------------------------------------*/

sort_population()
{
  int p, i, j, k;
  float best;
  
  /* Use insert sort */
  for ( p = 0 ; p < POPS ; p++ )
    {
      for ( i = 0 ; i < MAXPOP ; i++ )
	{
	  {
	    best = score[p][i];
	    for ( j = (i+1) ; j < MAXPOP ; j++ )
	      if ( score[p][j] > best )
		{
		  best = score[p][j];
		  k = j;
		}
	    if ( best > score[p][i] )
	      pop_swap( p, i, k);
	  }     
	}
    }
} 

/*-----------------------------------------------------------------------*\
|                                                                         |
|  irand( range )  - return a random integer in the range 0..(range-1)    |
|                                                                         |
\*-----------------------------------------------------------------------*/

int irand( range )
     int range;
{
  return( random() % range );
}

/*-----------------------------------------------------------------------*\
|                                                                         |
|  Uniform Crossover: pop[p1][i1][j] and pop[p2][i2][j] are swapped       |
|  randomly                                                               |
\*-----------------------------------------------------------------------*/

uniform_cross( p1, i1, p2, i2)
  int p1,i1,p2,i2;
{
  int j,k;
  float t;

  for( j=0; j < SIZE; j++)
    {
      k = irand(2);
      if( k==0 )
	{
	  t = pop[p1][i1][j];
	  pop[p1][i1][j] = pop[p2][i2][j];
	  pop[p2][i2][j] = t;
	}
    }
}
/*-----------------------------------------------------------------------*\
|                                                                         |
|   Show (on the standard output) the best scores of all populations      |
|                                                                         |
\*-----------------------------------------------------------------------*/

statistics( generation )
     int generation;
{
  int p,i;
  printf("generationi is here: %d\n",generation);
  if ( generation % MIXPOP == 0 )
    printf("-----------------------------\n");
  printf(" %4d) First are: ", generation);
  for ( p = 0 ; p < POPS ; p++ )  printf("%f ", score[p][0] );
  printf(" (from %d)\n",total);
}
  
/*-----------------------------------------------------------------------*\
|                                                                         |
|  Generate an intermediate generation                                    |
|                                                                         |
\*-----------------------------------------------------------------------*/

roulette_wheel()
{
  int p1, p2 = 0, i1, i2, j;
  int r, integer_part;
  double sum_score, sum_temp = 0; 
  
  for (p1 = 0; p1 < POPS; p1++ )
    {
      sum_score = 0;
      for( i1 = 0 ; i1 < MAXPOP ; i1++ )
	sum_score += score[p1][i1];
      integer_part = sum_score;
      
      for( i2 = 0 ; i2 < MAXPOP ; i2++ )
	{
	  i1 = 0;
	  r = random()&integer_part ;
	  sum_temp = score[p1][i1];
	  while (sum_temp <= r )
	    {
	      ++i1;
	      sum_temp += score[p1][i1];
	    }
	  for ( j = 0; j < SIZE ; j++ )
	   temp_pop[p2][i2][j] = pop[p1][i1][j];
	}
      ++p2;
    }
  for (p1 = 0; p1 < POPS; p1++ )
    for( i1 = 0 ; i1 < MAXPOP ; i1++ )
      for ( j = 0; j < SIZE ; j++ )
	pop[p1][i1][j]=temp_pop[p1][i1][j];
}

/*-----------------------------------------------------------------------*\
|                                                                         |
|  Generate the next generation in all populations                        |
|                                                                         |
\*-----------------------------------------------------------------------*/

make_next_generation( generation )
     int generation;
{
  int p, i, j, q, k, l, y[20];
  float dev, sum=0, depop_y[20], depop_fe;
  char c;
  FILE *afp;

  roulette_wheel();
  calc_score();
  sort_population();

  for ( p = 0 ; p < POPS ; p++ ) 
    {
      /* keep best - BESTPOP */
      
      /* crossover */
      for( i = BESTPOP; i < CROSSPOP ; i++)
	{
	  if((generation%MIXPOP)==0)
	    {
	      q=irand(POPS);
	      while(q==p)
		q=irand(POPS);
	      uniform_cross( p, i, q, irand(CROSSPOP - i) + i);
	    }
	  else
	    uniform_cross( p, i, p, irand(CROSSPOP - i) + i);
	}

      /* slight mutation */
      i = irand (CROSSPOP - BESTPOP)+ BESTPOP;
      {
	j = irand(SIZE);
	dev = 1 + ((irand(2000) - 1000)/500000.0);
	pop[p][i][j] *= dev;
	while( pop[p][i][j] < (-(sort_low[j]/(sort_high[j]-sort_low[j])+0.5)))
	  {	   
	    dev = 1 + ((irand(2000) - 1000)/500000.0);
	    pop[p][i][j] *= dev;
	  }
      }
      
      /* severe mutation */
      i = irand (CROSSPOP - BESTPOP)+ BESTPOP;
      {
	j = irand(SIZE);
	dev = 1 + ((irand(2000) - 1000)/5000.0);
	pop[p][i][j] *= dev;
	while( pop[p][i][j] < (-(sort_low[j]/(sort_high[j]-sort_low[j])+0.5)))
	  {
	    dev = 1 + ((irand(2000) - 1000)/5000.0);
	    pop[p][i][j] *= dev;
	  }   
	}
      
      /* create new individuals */
      for ( i = CROSSPOP ; i < MAXPOP ; i++ ) 
	{
	  sum = 0;	 
	  for ( k = 0 ; k < (SIZE-1) ; k++ )
	    {
	      y[k] = irand(SIZE-1);
	      for ( l = 0; l <k ; l++)
		{
		  while ( y[k] == y[l] )
		    {
		      y[k] = irand(SIZE-1);
		      l = 0;
		    }
		}
	      depop_y[k] = (random()&1000000) / 10000.0 ;
	      sum += depop_y[k];
	    }
	  depop_fe = (random()&1000000) / 10000.0 ;
	  sum += depop_fe;
	  for ( k = 0 ; k < (SIZE-1) ; k++ )
	    {
	      j = y[k];
	      pop[p][i][j] = ((depop_y[k]*100/sum)-(sort_low[j]))/(sort_high[j]-sort_low[j])-0.5;
	    }
	  pop[p][i][SIZE-1] = ((random()&1048575)/1000000.0 - 0.5) * 2;
	}
    }   
  
  calc_score();
  sort_population();
  statistics( generation );
  printf("Done %d generations\n\n",generation);
}

/*-----------------------------------------------------------------------*\
|                                                                         |
|   best_pop  -  Find the population with the best solution               |
|                                                                         |
\*-----------------------------------------------------------------------*/

int best_pop()
{
  int i, p;
  float best=0;
  for ( i = 0 ; i < POPS ; i++ )
    if ( score[i][0] > best )
      {
        best = score[i][0];
        p = i;
      }
  return(p);
}

/*-----------------------------------------------------------------------*\
|                                                                         |
|  Main                                                                   |
|                                                                         |
\*-----------------------------------------------------------------------*/

main() 
{
  int generation, i, j, l, p, done = 0, k,x1,x2,w,z;
  float y;
  FILE *fb,*fscore,*afp;
  char c[10],label[60]; 
  float value_Ti,value_Nb,value_C,value_N;

 /* read in the file "values" the numbers of inputs (NUM) and the        */    
 /* numbers of which are allowed to vary (SIZE)                          */
  if(( afp = fopen("values","r")) == NULL )
   {
     printf("error: can't read file \"values\"");
     exit(1);
   }
 else
   {
     fgets(label,sizeof(label),afp);     
     while(!feof(afp)) 
       {
	 fscanf(afp,"%s%d%d%f%f%f",c,&w,&z,&y,&y,&y);
	  if(z==1)
	    ++SIZE;
	  ++NUM;
       } 
   }
  fclose(afp);
 
 /* define the size of the array *f,*v,*low,*high,*sort_low,*sort_high   */    
 /* *p,***pop                                                            */
  f=(float*)malloc(NUM*sizeof(float));
  low=(float*)malloc(NUM*sizeof(float));
  high=(float*)malloc(NUM*sizeof(float));
  sort_low=(float*)malloc(SIZE*sizeof(float));
  sort_high=(float*)malloc(SIZE*sizeof(float));
  v=(int*)malloc(NUM*sizeof(int));
  x=(int*)malloc(NUM*sizeof(int));
  w1=(float*)malloc(SIZE*sizeof(float));
  pop=(float***)malloc(POPS*sizeof(float**));
  for(i = 0; i < POPS; i++)
    {
      pop[i]=(float**)malloc(MAXPOP*sizeof(float*));
      for(j = 0; j <  MAXPOP; j++)
	pop[i][j]=(float*)malloc(SIZE*sizeof(float));
    }
  temp_pop=(float***)malloc(POPS*sizeof(float**));
  for(i = 0; i < POPS; i++)
    {
     temp_pop[i]=(float**)malloc(MAXPOP*sizeof(float*));
      for(j = 0; j <  MAXPOP; j++)
	temp_pop[i][j]=(float*)malloc(SIZE*sizeof(float));
    }
  
  randomize();
  get_data();  /* Read input from file */
  calc_limit_err();
  make_initial_population();
  calc_score();
  sort_population();
  
  for (p=0;p<POPS;p++)
    {
      printf("Score of %d is %f",p,score[p][0]);
    }
  printf("We got here ");

  /* Educate the net */
  generation = 0;
  
  while ( (done != 1 ) && ( generation++ < SESSIONS ) )
    {    
      make_next_generation( generation );
      p = best_pop();      
      if ( score[p][0] > target_1)
	{
	  printf ("Target error was %f and error is %f",target_1,score[p][0]);
	  done = 1; 
	}
      else
	{
	  printf("Done all !!\n");
	}
    
      fb=fopen("../gacode/unnormalise/nn-output_b","w");
      for (i=0;i<POPS;i++) 
	{
	  for ( j=0;j<BESTPOP;j++) 
	    {     	     
	      apply (i,j);
	      l=0;	      
	      for (k=0;k<NUM;k++)
		{ 
		  if (v[k]==1)
		    {	   	  
		      fprintf(fb,"%f\n",w1[l]);
		      if (!strcmp(name[k],"Ti"))	
			value_Ti =  (w1[l]+0.5)*(high[k]-low[k])+low[k];
		      else
			{
			  if (!strcmp(name[k],"Nb"))	 
			    value_Nb = (w1[l]+0.5)*(high[k]-low[k])+low[k];
			  else
			    {	
			      if (!strcmp(name[k],"C"))  
				value_C = (w1[l]+0.5)*(high[k]-low[k])+low[k];
			      else
				{ 
				  if (!strcmp(name[k],"N")) 
				    value_N =  (w1[l]+0.5)*(high[k]-low[k])+low[k];
				}
			    }
			}
		      l++;
		    } 
		  else
		    {
		      if (v[k]==0)
			{  
			  fprintf(fb,"%f\n",(f[k]-low[k])/(high[k]-low[k])-0.5);
			  if (!strcmp(name[j],"Ti"))
			    value_Ti = f[k];
			  else
			    {
			      if (!strcmp(name[k],"Nb"))
				value_Nb = f[k];
			      else
				{
				  if (!strcmp(name[k],"C"))
				    value_C = f[k];
				  else
				    {		
				      if (!strcmp(name[k],"N"))
					value_N = f[k];
				    }
				}	
			    }
			}
		      else
			fprintf(fb,"%f\n",(((((value_Ti/4)+(value_Nb/8))/(value_C+value_N))-low[k])/(high[k]-low[k]))-0.5); 			 
		    }
		}
	      fprintf(fb,"%f %f  \n",result[i][j],(error[i][j]+result[i][j]));
	    }
	}
      fclose(fb);
    }  
      
  printf("done! \n");
  printf("\n\n****  For unnormalisation, goto /unnormalise and type a.out (no. of lines in output file) => BEST_POP x POPS = no of lines in output file  (no. of inputs)****\n");
  fscore=fopen("score_output","w"); 
  for (i=0;i<POPS;i++) { 
    for ( j=0;j<MAXPOP;j++) { 
      fprintf(fscore,"Score for score_T_298 pop %d gene %d is %f \n",i,j,score[i][j]); 
      printf("scores for score_T_298 pop %d chromo %d id %f \n",i,j,score[i][j]);
    } 
  }        
  printf("\n Scores printed to 'scores'\n\n");
  fclose(fscore);

  free(f);
  free(low);
  free(high);
  free(sort_low);
  free(sort_high);
  free(v);
  free(x);
  free(w1);
  free(pop);
  free(temp_pop);
}
